@@ -64,7 +64,7 @@ class RemoveAssignmentsToNames(VisitorTransform, SkipDeclarations):
     generated, while recording what it has removed so that it can be used in the initialization.
     """
     def __init__(self, names):
-        super(RemoveAssignmentsToNames, self).__init__()
+        super().__init__()
         self.names = names
         self.removed_assignments = {}
 
@@ -95,7 +95,7 @@ class RemoveAssignmentsToNames(VisitorTransform, SkipDeclarations):
         return node
 
 
-class TemplateCode(object):
+class TemplateCode:
     """
     Adds the ability to keep track of placeholder argument names to PyxCodeWriter.
 
@@ -112,9 +112,8 @@ class TemplateCode(object):
     def add_code_line(self, code_line):
         self.writer.putln(code_line)
 
-    def add_code_lines(self, code_lines):
-        for line in code_lines:
-            self.writer.putln(line)
+    def add_code_chunk(self, code_chunk):
+        self.writer.put_chunk(code_chunk)
 
     def reset(self):
         # don't attempt to reset placeholders - it really doesn't matter if
@@ -124,8 +123,14 @@ class TemplateCode(object):
     def empty(self):
         return self.writer.empty()
 
-    def indenter(self):
-        return self.writer.indenter()
+    def indent(self):
+        self.writer.indent()
+
+    def dedent(self):
+        self.writer.dedent()
+
+    def indenter(self, block_opener_line):
+        return self.writer.indenter(block_opener_line)
 
     def new_placeholder(self, field_names, value):
         name = self._new_placeholder_name(field_names)
@@ -139,7 +144,7 @@ class TemplateCode(object):
 
     def _new_placeholder_name(self, field_names):
         while True:
-            name = "DATACLASS_PLACEHOLDER_%d" % self._placeholder_count
+            name = f"DATACLASS_PLACEHOLDER_{self._placeholder_count:d}"
             if (name not in self.placeholders
                     and name not in field_names):
                 # make sure name isn't already used and doesn't
@@ -167,12 +172,12 @@ class TemplateCode(object):
         )
 
 
-class _MISSING_TYPE(object):
+class _MISSING_TYPE:
     pass
 MISSING = _MISSING_TYPE()
 
 
-class Field(object):
+class Field:
     """
     Field is based on the dataclasses.field class from the standard library module.
     It is used internally during the generation of Cython dataclasses to keep track
@@ -277,7 +282,7 @@ def process_class_get_fields(node):
                     # The standard library module generates a TypeError at runtime
                     # in this situation.
                     # Error message is copied from CPython
-                    error(assignment.pos, "mutable default <class '{0}'> for field {1} is not allowed: "
+                    error(assignment.pos, "mutable default <class '{}'> for field {} is not allowed: "
                           "use default_factory".format(assignment.type.name, name))
 
                 field = Field(node.pos, default=assignment)
@@ -296,7 +301,7 @@ def handle_cclass_dataclass(node, dataclass_args, analyse_decs_transform):
     # default argument values from https://docs.python.org/3/library/dataclasses.html
     kwargs = dict(init=True, repr=True, eq=True,
                   order=False, unsafe_hash=False,
-                  frozen=False, kw_only=False)
+                  frozen=False, kw_only=False, match_args=True)
     if dataclass_args is not None:
         if dataclass_args[0]:
             error(node.pos, "cython.dataclasses.dataclass takes no positional arguments")
@@ -327,7 +332,7 @@ def handle_cclass_dataclass(node, dataclass_args, analyse_decs_transform):
           for k, v in kwargs.items() ] +
         [ (ExprNodes.IdentifierStringNode(node.pos, value=EncodedString(k)),
            ExprNodes.BoolNode(node.pos, value=v))
-          for k, v in [('kw_only', kw_only), ('match_args', False),
+          for k, v in [('kw_only', kw_only),
                        ('slots', False), ('weakref_slot', False)]
         ])
     dataclass_params = make_dataclass_call_helper(
@@ -344,6 +349,7 @@ def handle_cclass_dataclass(node, dataclass_args, analyse_decs_transform):
 
     code = TemplateCode()
     generate_init_code(code, kwargs['init'], node, fields, kw_only)
+    generate_match_args(code, kwargs['match_args'], node, fields, kw_only)
     generate_repr_code(code, kwargs['repr'], node, fields)
     generate_eq_code(code, kwargs['eq'], node, fields)
     generate_order_code(code, kwargs['order'], node, fields)
@@ -398,6 +404,7 @@ def generate_init_code(code, init, node, fields, kw_only):
 
     function_start_point = code.insertion_point()
     code = code.insertion_point()
+    code.indent()
 
     # create a temp to get _HAS_DEFAULT_FACTORY
     dataclass_module = make_dataclasses_module_callnode(node.pos)
@@ -413,17 +420,18 @@ def generate_init_code(code, init, node, fields, kw_only):
     for name, field in fields.items():
         entry = node.scope.lookup(name)
         if entry.annotation:
-            annotation = u": %s" % entry.annotation.string.value
+            annotation = f": {entry.annotation.string.value}"
         else:
-            annotation = u""
-        assignment = u''
+            annotation = ""
+        assignment = ''
         if field.default is not MISSING or field.default_factory is not MISSING:
-            seen_default = True
+            if field.init.value:
+                seen_default = True
             if field.default_factory is not MISSING:
                 ph_name = default_factory_placeholder
             else:
                 ph_name = code.new_placeholder(fields, field.default)  # 'default' should be a node
-            assignment = u" = %s" % ph_name
+            assignment = f" = {ph_name}"
         elif seen_default and not kw_only and field.init.value:
             error(entry.pos, ("non-default argument '%s' follows default argument "
                               "in dataclass __init__") % name)
@@ -431,39 +439,58 @@ def generate_init_code(code, init, node, fields, kw_only):
             return
 
         if field.init.value:
-            args.append(u"%s%s%s" % (name, annotation, assignment))
+            args.append(f"{name}{annotation}{assignment}")
 
         if field.is_initvar:
             continue
         elif field.default_factory is MISSING:
             if field.init.value:
-                code.add_code_line(u"    %s.%s = %s" % (selfname, name, name))
+                code.add_code_line(f"{selfname}.{name} = {name}")
             elif assignment:
                 # not an argument to the function, but is still initialized
-                code.add_code_line(u"    %s.%s%s" % (selfname, name, assignment))
+                code.add_code_line(f"{selfname}.{name}{assignment}")
         else:
             ph_name = code.new_placeholder(fields, field.default_factory)
             if field.init.value:
                 # close to:
                 # def __init__(self, name=_PLACEHOLDER_VALUE):
                 #     self.name = name_default_factory() if name is _PLACEHOLDER_VALUE else name
-                code.add_code_line(u"    %s.%s = %s() if %s is %s else %s" % (
-                    selfname, name, ph_name, name, default_factory_placeholder, name))
+                code.add_code_line(
+                    f"{selfname}.{name} = {ph_name}() if {name} is {default_factory_placeholder} else {name}"
+                )
             else:
                 # still need to use the default factory to initialize
-                code.add_code_line(u"    %s.%s = %s()" % (
-                    selfname, name, ph_name))
+                code.add_code_line(f"{selfname}.{name} = {ph_name}()")
 
     if node.scope.lookup("__post_init__"):
         post_init_vars = ", ".join(name for name, field in fields.items()
                                    if field.is_initvar)
-        code.add_code_line("    %s.__post_init__(%s)" % (selfname, post_init_vars))
+        code.add_code_line(f"{selfname}.__post_init__({post_init_vars})")
 
     if code.empty():
-        code.add_code_line("    pass")
+        code.add_code_line("pass")
 
-    args = u", ".join(args)
-    function_start_point.add_code_line(u"def __init__(%s):" % args)
+    args = ", ".join(args)
+    function_start_point.add_code_line(f"def __init__({args}):")
+
+
+def generate_match_args(code, match_args, node, fields, global_kw_only):
+    """
+    Generates a tuple containing what would be the positional args to __init__
+
+    Note that this is generated even if the user overrides init
+    """
+    if not match_args or node.scope.lookup_here("__match_args__"):
+        return
+    positional_arg_names = []
+    for field_name, field in fields.items():
+        # TODO hasattr and global_kw_only can be removed once full kw_only support is added
+        field_is_kw_only = global_kw_only or (
+            hasattr(field, 'kw_only') and field.kw_only.value
+        )
+        if not field_is_kw_only:
+            positional_arg_names.append(field_name)
+    code.add_code_line("__match_args__ = %s" % str(tuple(positional_arg_names)))
 
 
 def generate_repr_code(code, repr, node, fields):
@@ -500,25 +527,35 @@ def generate_repr_code(code, repr, node, fields):
             break
 
     if needs_recursive_guard:
-        code.add_code_line("__pyx_recursive_repr_guard = __import__('threading').local()")
-        code.add_code_line("__pyx_recursive_repr_guard.running = set()")
-    code.add_code_line("def __repr__(self):")
-    if needs_recursive_guard:
-        code.add_code_line("    key = id(self)")
-        code.add_code_line("    guard_set = self.__pyx_recursive_repr_guard.running")
-        code.add_code_line("    if key in guard_set: return '...'")
-        code.add_code_line("    guard_set.add(key)")
-        code.add_code_line("    try:")
-    strs = [u"%s={self.%s!r}" % (name, name)
-            for name, field in fields.items()
-            if field.repr.value and not field.is_initvar]
-    format_string = u", ".join(strs)
+        code.add_code_chunk("""
+            __pyx_recursive_repr_guard = __import__('threading').local()
+            __pyx_recursive_repr_guard.running = set()
+        """)
 
-    code.add_code_line(u'        name = getattr(type(self), "__qualname__", type(self).__name__)')
-    code.add_code_line(u"        return f'{name}(%s)'" % format_string)
-    if needs_recursive_guard:
-        code.add_code_line("    finally:")
-        code.add_code_line("        guard_set.remove(key)")
+    with code.indenter("def __repr__(self):"):
+        if needs_recursive_guard:
+            code.add_code_chunk("""
+                key = id(self)
+                guard_set = self.__pyx_recursive_repr_guard.running
+                if key in guard_set: return '...'
+                guard_set.add(key)
+                try:
+            """)
+            code.indent()
+
+        strs = ["%s={self.%s!r}" % (name, name)
+                for name, field in fields.items()
+                if field.repr.value and not field.is_initvar]
+        format_string = ", ".join(strs)
+
+        code.add_code_chunk(f'''
+            name = getattr(type(self), "__qualname__", None) or type(self).__name__
+            return f'{{name}}({format_string})'
+        ''')
+        if needs_recursive_guard:
+            code.dedent()
+            with code.indenter("finally:"):
+                code.add_code_line("guard_set.remove(key)")
 
 
 def generate_cmp_code(code, op, funcname, node, fields):
@@ -527,39 +564,33 @@ def generate_cmp_code(code, op, funcname, node, fields):
 
     names = [name for name, field in fields.items() if (field.compare.value and not field.is_initvar)]
 
-    code.add_code_lines([
-        "def %s(self, other):" % funcname,
-        "    if other.__class__ is not self.__class__:"
-        "        return NotImplemented",
+    with code.indenter(f"def {funcname}(self, other):"):
+        code.add_code_chunk(f"""
+            if other.__class__ is not self.__class__: return NotImplemented
+
+            cdef {node.class_name} other_cast
+            other_cast = <{node.class_name}>other
+        """)
+
+        # The Python implementation of dataclasses.py does a tuple comparison
+        # (roughly):
+        #  return self._attributes_to_tuple() {op} other._attributes_to_tuple()
         #
-        "    cdef %s other_cast" % node.class_name,
-        "    other_cast = <%s>other" % node.class_name,
-    ])
+        # For the Cython implementation a tuple comparison isn't an option because
+        # not all attributes can be converted to Python objects and stored in a tuple
+        #
+        # TODO - better diagnostics of whether the types support comparison before
+        #    generating the code. Plus, do we want to convert C structs to dicts and
+        #    compare them that way (I think not, but it might be in demand)?
+        checks = []
+        op_without_equals = op.replace('=', '')
 
-    # The Python implementation of dataclasses.py does a tuple comparison
-    # (roughly):
-    #  return self._attributes_to_tuple() {op} other._attributes_to_tuple()
-    #
-    # For the Cython implementation a tuple comparison isn't an option because
-    # not all attributes can be converted to Python objects and stored in a tuple
-    #
-    # TODO - better diagnostics of whether the types support comparison before
-    #    generating the code. Plus, do we want to convert C structs to dicts and
-    #    compare them that way (I think not, but it might be in demand)?
-    checks = []
-    op_without_equals = op.replace('=', '')
-
-    for name in names:
-        if op != '==':
-            # tuple comparison rules - early elements take precedence
-            code.add_code_line("    if self.%s %s other_cast.%s: return True" % (
-                name, op_without_equals, name))
-        code.add_code_line("    if self.%s != other_cast.%s: return False" % (
-            name, name))
-    if "=" in op:
-        code.add_code_line("    return True")  # "() == ()" is True
-    else:
-        code.add_code_line("    return False")
+        for name in names:
+            if op != '==':
+                # tuple comparison rules - early elements take precedence
+                code.add_code_line(f"if self.{name} {op_without_equals} other_cast.{name}: return True")
+            code.add_code_line(f"if self.{name} != other_cast.{name}: return False")
+        code.add_code_line(f"return {'True' if '=' in op else 'False'}")  # "() == ()" is True
 
 
 def generate_eq_code(code, eq, node, fields):
@@ -646,15 +677,13 @@ def generate_hash_code(code, unsafe_hash, eq, frozen, node, fields):
     ]
 
     # make a tuple of the hashes
-    hash_tuple_items = u", ".join(u"self.%s" % name for name in names)
+    hash_tuple_items = ", ".join("self.%s" % name for name in names)
     if hash_tuple_items:
-        hash_tuple_items += u","  # ensure that one arg form is a tuple
+        hash_tuple_items += ","  # ensure that one arg form is a tuple
 
     # if we're here we want to generate a hash
-    code.add_code_lines([
-        "def __hash__(self):",
-        "    return hash((%s))" % hash_tuple_items,
-    ])
+    with code.indenter("def __hash__(self):"):
+        code.add_code_line(f"return hash(({hash_tuple_items}))")
 
 
 def get_field_type(pos, entry):
@@ -680,7 +709,7 @@ def get_field_type(pos, entry):
         #)
         #return ExprNodes.IndexNode(
         #    pos, base=annotations,
-        #    index=ExprNodes.StringNode(pos, value=entry.name)
+        #    index=ExprNodes.UnicodeNode(pos, value=entry.name)
         #)
     else:
         # it's slightly unclear what the best option is here - we could
@@ -688,7 +717,7 @@ def get_field_type(pos, entry):
         # attributes defined with cdef so Cython is free to make it's own
         # decision
         s = EncodedString(entry.type.declaration_code("", for_display=1))
-        return ExprNodes.StringNode(pos, value=s)
+        return ExprNodes.UnicodeNode(pos, value=s)
 
 
 class FieldRecordNode(ExprNodes.ExprNode):
@@ -703,7 +732,7 @@ class FieldRecordNode(ExprNodes.ExprNode):
     subexprs = ['arg']
 
     def __init__(self, pos, arg):
-        super(FieldRecordNode, self).__init__(pos, arg=arg)
+        super().__init__(pos, arg=arg)
 
     def analyse_types(self, env):
         self.arg.analyse_types(env)
@@ -723,7 +752,7 @@ class FieldRecordNode(ExprNodes.ExprNode):
         from .AutoDocTransforms import AnnotationWriter
         writer = AnnotationWriter(description="Dataclass field")
         string = writer.write(self.arg)
-        return ExprNodes.StringNode(self.pos, value=EncodedString(string))
+        return ExprNodes.UnicodeNode(self.pos, value=EncodedString(string))
 
     def generate_evaluation_code(self, code):
         return self.arg.generate_evaluation_code(code)
@@ -816,11 +845,11 @@ def _set_up_dataclass_fields(node, fields, dataclass_module):
                 key=ExprNodes.IdentifierStringNode(node.pos, value=EncodedString(name)),
                 value=dc_field_call))
         dc_fields_namevalue_assignments.append(
-            dedent(u"""\
-                __dataclass_fields__[{0!r}].name = {0!r}
-                __dataclass_fields__[{0!r}].type = {1}
-                __dataclass_fields__[{0!r}]._field_type = {2}
-            """).format(name, type_placeholder_name, field_type_placeholder_name))
+            dedent(f"""\
+                __dataclass_fields__[{name!r}].name = {name!r}
+                __dataclass_fields__[{name!r}].type = {type_placeholder_name}
+                __dataclass_fields__[{name!r}]._field_type = {field_type_placeholder_name}
+            """))
 
     dataclass_fields_assignment = \
         Nodes.SingleAssignmentNode(node.pos,
@@ -828,7 +857,7 @@ def _set_up_dataclass_fields(node, fields, dataclass_module):
                                         name=EncodedString("__dataclass_fields__")),
                         rhs = dc_fields)
 
-    dc_fields_namevalue_assignments = u"\n".join(dc_fields_namevalue_assignments)
+    dc_fields_namevalue_assignments = "\n".join(dc_fields_namevalue_assignments)
     dc_fields_namevalue_assignments = TreeFragment(dc_fields_namevalue_assignments,
                                                    level="c_class",
                                                    pipeline=[NormalizeTree(None)])
